package ru.dbotthepony.mc.otm.entity
import net.minecraft.sounds.SoundEvent
import net.minecraft.sounds.SoundEvents
import net.minecraft.world.entity.*
import net.minecraft.world.entity.ai.attributes.AttributeSupplier
import net.minecraft.world.entity.ai.attributes.Attributes
import net.minecraft.world.entity.ai.control.MoveControl
import net.minecraft.world.entity.ai.goal.*
import net.minecraft.world.entity.ai.goal.target.HurtByTargetGoal
import net.minecraft.world.entity.ai.goal.target.NearestAttackableTargetGoal
import net.minecraft.world.entity.ai.navigation.FlyingPathNavigation
import net.minecraft.world.entity.ai.navigation.PathNavigation
import net.minecraft.world.entity.monster.Monster
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.Level
import net.minecraft.world.phys.Vec3
import ru.dbotthepony.mc.otm.registry.game.MEntityTypes
import java.util.*
import kotlin.math.cos
import kotlin.math.sin

class Drone(type: EntityType<Drone>, level: Level) : Monster(type, level) {
	constructor(level: Level) : this(MEntityTypes.DRONE, level)

	var attackTarget: LivingEntity? = null
	var attackTime = 0
	var attackDuration = 40

	val idleState = AnimationState()

	override fun createNavigation(level: Level): PathNavigation = FlyingPathNavigation(this, level())

	init {
	    idleState.start(tickCount)
		this.isNoGravity = true
	}

	override fun registerGoals() {
		goalSelector.addGoal(8, RandomLookAroundGoal(this))

		goalSelector.addGoal(8, LookAtPlayerGoal(this, Player::class.java, 12f))

		goalSelector.addGoal(3, NearestAttackableTargetGoal(this, Player::class.java , true, true))

		goalSelector.addGoal(2, DroneBeamAttackGoal(this))

		targetSelector.addGoal(1, HurtByTargetGoal(this))
	}


	override fun getHurtSound(damageSource: net.minecraft.world.damagesource.DamageSource): SoundEvent {
		return SoundEvents.HEAVY_CORE_BREAK
	}

	override fun getDeathSound(): SoundEvent {
		return SoundEvents.VAULT_BREAK
	}

	//essentially same attack as guardians
	//probably needs to render a beam too
	class DroneBeamAttackGoal(private val mob: Drone) : Goal() {
		private var attackCooldown = 0

		override fun canUse(): Boolean {
			val target = mob.target
			return target != null && mob.distanceToSqr(target) < 100
		}

		override fun start() {
			mob.attackTarget = mob.target
			mob.attackTime = 0
		}

		override fun stop() {
			mob.attackTarget = null
			mob.attackTime = 0
		}

		override fun tick() {
			val target = mob.attackTarget ?: return

			mob.lookAt(target, 30.0f, 30.0f)

			if (mob.attackTime < mob.attackDuration) {
				mob.attackTime++
				if (mob.attackTime == mob.attackDuration / 2) {
					mob.playSound(SoundEvents.GUARDIAN_ATTACK, 1.0f, 1.0f)
				}
			} else {
				val direction = Vec3(target.x - mob.x, target.eyeY - mob.eyeY, target.z - mob.z).normalize()
				target.hurt(mob.damageSources().magic(), 6.0f)

				mob.attackTime = 0
				attackCooldown = 60
			}

			if (attackCooldown > 0) {
				attackCooldown--
			}
		}
	}

	companion object {
		fun createAttributes() : AttributeSupplier.Builder {
			return createMonsterAttributes()
				.add(Attributes.MAX_HEALTH, 10.0)
				.add(Attributes.MOVEMENT_SPEED, 0.4)
				.add(Attributes.ATTACK_DAMAGE, 3.0)
				.add(Attributes.GRAVITY, 0.0)
				.add(Attributes.KNOCKBACK_RESISTANCE, 0.4)
				.add(Attributes.FALL_DAMAGE_MULTIPLIER, 0.0)
		}
	}
}
