package ru.dbotthepony.mc.otm.menu.tech

import net.minecraft.core.Direction
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.level.block.Block
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.mc.otm.block.entity.RedstoneSetting
import ru.dbotthepony.mc.otm.block.tech.EnergyCounterBlock
import ru.dbotthepony.mc.otm.block.entity.tech.EnergyCounterBlockEntity
import ru.dbotthepony.mc.otm.util.chart.DecimalHistoryChart
import ru.dbotthepony.mc.otm.util.math.Decimal
import ru.dbotthepony.mc.otm.util.math.toDecimal
import ru.dbotthepony.mc.otm.menu.MatteryMenu
import ru.dbotthepony.mc.otm.menu.input.BooleanInputWithFeedback
import ru.dbotthepony.mc.otm.menu.input.EnumInputWithFeedback
import ru.dbotthepony.mc.otm.menu.input.IntInputWithFeedback
import ru.dbotthepony.mc.otm.registry.game.MMenus
import java.math.BigDecimal
import java.util.function.Supplier

class EnergyCounterMenu(
	p_38852_: Int,
	inventory: Inventory,
	tile: EnergyCounterBlockEntity? = null
) : MatteryMenu(MMenus.ENERGY_COUNTER, p_38852_, inventory, tile) {
	var passed by mSynchronizer.computedDecimal { tile?.passed ?: Decimal.ZERO }
	var lastTick by mSynchronizer.computedDecimal { tile?.lastTick ?: Decimal.ZERO }
	var maxIO by mSynchronizer.computedDecimal { tile?.ioLimit ?: Decimal.ZERO }

	val history5s = tile?.history5s ?: DecimalHistoryChart(1, 100)
	val history15s = tile?.history15s ?: DecimalHistoryChart(3, 100)
	val history1m = tile?.history1m ?: DecimalHistoryChart(12, 100)
	val history10m = tile?.history10m ?: DecimalHistoryChart(120, 100)
	val history1h = tile?.history1h ?: DecimalHistoryChart(720, 100)
	val history6h = tile?.history6h ?: DecimalHistoryChart(720 * 6, 100)
	val history24h = tile?.history24h ?: DecimalHistoryChart(720 * 24, 100)

	val pullEnergyFromInput = BooleanInputWithFeedback(this, tile?.let { it::pullEnergyFromInput })

	init {
		mSynchronizer.add(history5s)
		mSynchronizer.add(history15s)
		mSynchronizer.add(history1m)
		mSynchronizer.add(history10m)
		mSynchronizer.add(history1h)
		mSynchronizer.add(history6h)
		mSynchronizer.add(history24h)
	}

	val switchDirection = oneWayInput {
		if (tile is EnergyCounterBlockEntity)
			tile.level?.setBlock(tile.blockPos, tile.blockState.setValue(EnergyCounterBlock.INPUT_DIRECTION, tile.blockState.getValue(EnergyCounterBlock.INPUT_DIRECTION).opposite), Block.UPDATE_ALL)
	}

	val displayChartOnBlock = IntInputWithFeedback(this)
	val redstone = EnumInputWithFeedback(this, RedstoneSetting::class.java)

	init {
		if (tile != null) {
			redstone.with(tile.redstoneControl::redstoneSetting)
			displayChartOnBlock.with(tile::displayChartOnBlock)
		}
	}

	val inputDirection: Direction by mSynchronizer.computedEnum { tile?.blockState?.getValue(EnergyCounterBlock.INPUT_DIRECTION) ?: Direction.UP}

	val maxIOInput = decimalInput {
		if (tile is EnergyCounterBlockEntity) {
			if (it.signum() < 0) {
				tile.ioLimit = null
			} else {
				tile.ioLimit = it
			}
		}
	}
}
